# Thesis Project - VMS Console
# File: DbFunctions.py
#
# Description: Provides functions, constants for DB functionality.
#
# Author: David Gallegos
# Last Updated: 08/30/2014

import os		#Provides directory functions
import xlrd 	#Excel Read - Python-Excel.org
import datetime #For manipulating datetime objects
import psycopg2 #Postgre driver

from Path  import * #Supporting class that logically clusters data points.
from Point import * #Supporting class that contains the fields of a GPS record.

#Configure to target VMS Postgres database
conn_string = "host='localhost' dbname='vms' user='postgres' password='password'"

#column numbers for Excel import
ASSET_ID = 1
SENT_DATE_UTC = 2
LATITUDE = 3
LONGITUDE = 4


def import_data_from_excel(file_path = ""):
    '''Imports Excel data, connects to DB and inserts all information.'''
    print("Connecting to database...")
    conn = psycopg2.connect(conn_string)
    cursor = conn.cursor()
    print("Connected.")
    
    #Various SQL statements that may be run depending on input.
    path_insertion = "INSERT INTO vessel_path (vessel_id, vessel_path_start) VALUES (%s, %s) RETURNING vessel_path_id"
    point_insertion = "INSERT INTO vessel_point (vessel_path_id, point_timestamp, vessel_behavior_id, coords, duration_in_seconds, speed, net_speed, movement_angle) VALUES (%s, %s, %s, ST_GeomFromText('POINT(%s %s)', 4269), %s, %s, %s, %s)"
    
    if file_path == "":
        file_path = input("Input path to Excel VMS data: ")
    print("Path: " + file_path)
    
    exists = os.path.exists(file_path)
    if not exists:
        print("Path does not exist. Please try again.")
        return

    print("Loading file -- This may take a while.")
    book = xlrd.open_workbook(file_path)

    print("Number of sheets = %d" % book.nsheets)

    # Dictionary where the key is the vessel,
    # and the values are the data points associated with that vessel
    vms_data = dict()

    # Iterate through each sheet in the workbook
    # TODO: Get the titles of the worksheets
    for index in range(book.nsheets):
        sheet = book.sheet_by_index(index)
        print("Starting to read sheet %d" % index)
        if (sheet.cell(0,ASSET_ID).value != 'ASSET_ID'):
            print('Error: ASSET_ID must be second column in spreadsheet')
            return
        elif (sheet.cell(0,SENT_DATE_UTC).value != 'SENT_DATE_UTC'):
            print('Error: SENT_DATE_UTC must be third column in spreadsheet')
            return
        elif (sheet.cell(0,LATITUDE).value != 'LATITUDE'):
            print('Error: LATITUDE must be fourth column in spreadsheet')
            return
        elif (sheet.cell(0,LONGITUDE).value != 'LONGITUDE'):
            print('Error: LONGITUDE must be fifth column in spreadsheet')
            return
        else:
            print('Spreadsheet format OK.')

        print("Interpreting data.")
        for rownum in range(1, sheet.nrows):
            asset_id = int(sheet.cell(rownum, ASSET_ID).value)

            #Construct new point object from data in row
            p = Point(
                excel_date_to_datetime(sheet.cell(rownum, SENT_DATE_UTC).value),
                sheet.cell(rownum, LATITUDE).value,
                sheet.cell(rownum, LONGITUDE).value
            )

            if not asset_id in vms_data:
                #Asset ID has not been used before, add a new one.
                vms_data[asset_id] = Path()

            #Append the new data point
            vms_data[asset_id].add_data(p)

        print("SIZE = " + str(len(vms_data)))
    print(file_path + ": Data read finished.")

    for asset_id, path in vms_data.items():
        trips = path.get_trips()
        print("Writing %d trips for vessel %d." % (len(trips), asset_id))

        for t in trips:
            #Insert vessel ID, first date from the trip.
            try:
                cursor.execute(path_insertion, (asset_id, t[0].timestamp))

                #obtain inserted path id
                vessel_path_id = cursor.fetchone()[0]

                #For each point in this path (continuous set of points closely clustered together),
                #perform an insertion operation.
                for p in t:
                    cursor.execute(point_insertion, (vessel_path_id, p.timestamp, p.behavior, p.longitude, p.latitude, p.duration, p.speed, p.net_speed, p.movement_angle))
                conn.commit()
            except psycopg2.IntegrityError:
                #Occurs if the path_insertion wasn't successful because of a constraint failure,
                #So ignore this record by rolling back because it's already in the system.
                print("Warning: Ignoring trip starting at %s, already exists in system for this vessel." % t[0].timestamp)
                conn.rollback()

    cursor.close()
    conn.close()
    
    print("Finished writing to database")


def delete_all_data():
    '''Deletes everything from record tables.'''
    deletion_query = "TRUNCATE vessel_point, vessel_path RESTART IDENTITY"

    delete_old = input('Are you sure you want to delete existing database information? (y/n): ')
    if (str.lower(delete_old) == 'y'):
        print("Connecting to database...")
        conn = psycopg2.connect(conn_string)
        cursor = conn.cursor()
        print("Connected. Deleting all data...")

        cursor.execute(deletion_query)
        print("Deletion has completed.")

    conn.commit()
    cursor.close()
    conn.close()


def db_fetch(query, args=None, single=False):
    '''Run query and get one result (like an aggregation). Can accept a list of queries.'''
    
    if isinstance(query, (list, tuple)):
        #Argument is a list of queries
        responses = []
        
        for single_query in query:
            #For each query in list, recursively call this function.
            responses += db_fetch(single_query, args, single)

        #Return array of responses, will maintain its original order.
        return responses
    else:
        #Single query
        conn = psycopg2.connect(conn_string)
        cursor = conn.cursor()

        cursor.execute(query, args)
        response = cursor.fetchall()
        if single:
            response = response[0]

        cursor.close()
        conn.close()

        return response


def excel_date_to_datetime(excel_date):
    '''Converts an excel float date to a Python date.'''
    return datetime.datetime(1900, 1, 1) + datetime.timedelta(days=excel_date - 2) #The -2 is a known workaround in Excel dates


def get_filters():
    '''Provides a common implementation to get filters on vessel and date ranges, if any.'''
    vessel = ""
    start_date = ""
    end_date = ""
    
    f = input("Filter by vessel ID? (y/n): ")
    while f.upper() == 'Y' and not isinstance(vessel, int):
        try:
            vessel = int(input("Enter vessel ID to filter: "))
        except ValueError:
            print("Incorrect format, must be a number.")
    f = input("Filter by date range? (y/n): ")
    while f.upper() == 'Y' and not isinstance(start_date, datetime.datetime):
        i = str(input('Input start date (mm/dd/yyyy): '))
        try:
            start_date = datetime.datetime.strptime(i, '%m/%d/%Y')
        except ValueError:
            print("Incorrect date format, use mm/dd/yyyy.")
    while f.upper() == 'Y' and not isinstance(end_date, datetime.datetime):
        i = str(input('Input end date (mm/dd/yyyy): '))
        try:
            end_date = datetime.datetime.strptime(i, '%m/%d/%Y')
        except ValueError:
            print("Incorrect date format, use mm/dd/yyyy.")
    return (vessel, start_date, end_date)


def build_query(query, vessel, start_date, end_date, post_where = "", date_name = "point_timestamp"):
    '''Common implementation to apply filters to query using vessel ID and date range.'''
    where_added = False
    args = []
    
    if vessel:
        where_added = True
        query += " WHERE vessel_id = %s"
        args = [vessel]
    if start_date:
        query += (" AND " if where_added else " WHERE ") + date_name + " >= %s AND " + date_name + " < %s"
        
        #It is safer in SQL to compare a date range by adding one to the end and comparing to less than that. 
        #This ensures that all records are caught, even those whose timestamp places them late in the day.
        args += [start_date, end_date + datetime.timedelta(days=1)]
    query += " " + post_where
    return (query, tuple(args))


def run_report_query(query, post_where = "", date_name = "point_timestamp"):
    '''Retrieves arguments, builds a query from its base (without WHERE), executes query and returns array of output.'''
    #prompt user if they want to filter data by criteria.
    vessel, start_date, end_date = get_filters()
    
    #build the query based on the filters.
    query, args = build_query(query, vessel, start_date, end_date, post_where, date_name)
    
    #query the database with query using obtained arguments.
    return db_fetch(query, args)
