# Thesis Project - SVM Behavior Categorization
# Author - Devin Frey

import matplotlib.pyplot as plt  # For graphing data
from sklearn.metrics import roc_curve, auc  # For graphing roc curves
from sklearn.preprocessing import label_binarize  # For pre-processing ROC data
import numpy  # For ravel function
from openpyxl import load_workbook  # For reading xlsx documents


# Helper method to produce a 2d graph of points used to train the SVM
def graph_classifier_predictions(X_test, y_test):
    plt.grid()
    ax = plt.gca()
    ax.set_axis_bgcolor((.5, .5, .5))
    ax.set_title(
        'Vessel Behavior as Classified by Deterministic Classifier'
    )

    d = {0: [], 1: [], 2: [], 3: [], 4: []}
    labels = ["Unknown", "Docked", "Fishing", "Traveling", "Poling"]
    for index, value in enumerate(y_test):
        d[value].append(index)
    colors = ['b', 'c', 'y', 'm', 'r']
    for key, indices in enumerate(d):
        stuff = d[key]
        x_stuff = []
        y_stuff = []
        for index in stuff:
            x_stuff.append(X_test[index, 0])
            y_stuff.append(X_test[index, 1])
        plt.scatter(x_stuff, y_stuff, color=colors[key], label=labels[key])
    plt.legend()
    plt.xlabel("speed")
    plt.ylabel("net speed")
    # plt.xticks(numpy.arange(-1, 1.25, .25))
    # plt.yticks(numpy.arange(-1, 1.25, .25))
    plt.show()


# Graph the original classifier side-by-side with the new classifier
def graph_original_vs_svm(clf, score, X_test, y_test):
    f, (ax1, ax2) = plt.subplots(1, 2, sharex=True, sharey=True)

    ax1.grid()
    ax1.set_axis_bgcolor((.5, .5, .5))
    ax1.set_title(
        'Original Classifier'
    )

    ax2.grid()
    ax2.set_axis_bgcolor((.5, .5, .5))
    ax2.set_title(
        'New Classifier, ' + str(score*100) + '% Accurate'
    )

    # A dictionary of behavior values
    d = {0: [], 1: [], 2: [], 3: [], 4: []}
    labels = ["Unknown", "Docked", "Fishing", "Traveling", "Poling"]
    # Put every index into the dictionary based on its behavior
    for index, value in enumerate(y_test):
        d[value].append(index)
    # The colors used in our scatter plot
    colors = ['b', 'c', 'y', 'm', 'r']

    # for behavior in our dictionary,
    for key, indices in enumerate(d):
        # gather all the indices of the points with behavior == key
        stuff = d[key]
        x_stuff = []
        y_stuff = []
        # for each point index, save its speed in x_stuff and its movement_angle in y_stuff
        for index in stuff:
            x_stuff.append(X_test[index, 0])
            y_stuff.append(X_test[index, 1])
        # Once all speeds/movement_angles have been cataloged for a behavior,
        # plot those points on our scatter plot in their appropriate color
        ax1.scatter(x_stuff, y_stuff, color=colors[key], label=labels[key])

    # Draw the legend
    ax1.legend()

    for index, value in enumerate(clf.predict(X_test)):
        d[value].append(index)
    colors = ['b', 'c', 'y', 'm', 'r']

    for key, indices in enumerate(d):
        stuff = d[key]
        x_stuff = []
        y_stuff = []
        for index in stuff:
            x_stuff.append(X_test[index, 0])
            y_stuff.append(X_test[index, 1])
        ax2.scatter(x_stuff, y_stuff, color=colors[key], label=labels[key])

    ax2.legend()

    f.set_size_inches(20, 11, forward=True)
    plt.show()
    # f.savefig(
    #     'figures/%s/Fig_K-%s_C-%d_G-%s_WS-%d_#P-%d.png'
    #     % (clf_params['kernel'], clf_params['kernel'], clf_params['C'], clf_params['gamma'], window_size, number_of_paths),
    #     dpi=300
    # )


def graph_windowing_accuracy(start_index, end_index):
    wb = load_workbook(filename='output.xlsx')
    ws = wb.active

    window_sizes = []
    accuracies = []
    for i in range(start_index, end_index):
        window_sizes.append(ws['E' + str(i)].value)
        accuracies.append(ws['H' + str(i)].value)
    plt.plot(window_sizes, accuracies, 'r--')
    plt.xlabel('Window Size')
    plt.ylabel('Accuracy')
    plt.title('Window Size vs. Accuracy')
    plt.grid()
    plt.show()


# Method to plot a line graph. Currently configured to plot accuracy vs window size for scaled and unscaled data sets.
def versus_line_graph(lines={"Scaled": {'start_index': 1517, 'end_index': 1566}, "Unscaled": {'start_index': 1567, 'end_index': 1616}}, x_col='E', y_col='H', x_name='Window Size', y_name='Accuracy'):
    wb = load_workbook(filename='output.xlsx')
    ws = wb.active
    line_colors = ['r--', 'b--', 'y--']

    # Iterate through each line of our graph (there are two in this case: scaled and unscaled)
    for index, key in enumerate(lines.keys()):
        line = lines[key]
        x_dim = []
        y_dim = []
        # Iterate through each row in the line's range
        for i in range(line['start_index'], line['end_index']):
            # Gather the values of our x and y dimensions
            x_dim.append(ws[x_col + str(i)].value)
            y_dim.append(ws[y_col + str(i)].value)
        # Plot the gathered data
        plt.plot(x_dim, y_dim, line_colors[index], label=key)

    # Label the graph then show it
    plt.xlabel(x_name)
    plt.ylabel(y_name)
    plt.title(x_name + ' vs. ' + y_name + " (Linear Kernel, C=1000, Jul2013-Aug2014, First 50 Paths with Over 50 Points, Docked Points, and No Unknowns)")
    plt.legend(loc="lower right")
    plt.grid()
    plt.show()


# Method to display the effect of C and gamma on an RBF SVM classifier
def svm_parameter_visualization(x, y, classifiers):
    C_2d_range = [1e-2, 1, 1e2]
    gamma_2d_range = [1e-1, 1, 1e1]
    plt.figure(figsize=(8, 6))
    xx, yy = numpy.meshgrid(numpy.linspace(-1.1, 1.1, 200), numpy.linspace(-1.1, 1.1, 200))
    for (k, (C, gamma, clf)) in enumerate(classifiers):
        # evaluate decision function in a grid
        Z = clf.decision_function(numpy.c_[xx.ravel(), yy.ravel()])
        Z = Z.reshape(xx.shape)

        # visualize decision function for these parameters
        plt.subplot(len(C_2d_range), len(gamma_2d_range), k + 1)
        plt.title("gamma=10^%d, C=10^%d" % (numpy.log10(gamma), numpy.log10(C)),
                  size='medium')

        # visualize parameter's effect on decision function
        plt.pcolormesh(xx, yy, -Z, cmap=plt.cm.autumn)
        plt.scatter(x[:, 0], x[:, 1], c=y, cmap=plt.cm.autumn_r)
        plt.xticks(())
        plt.yticks(())
        plt.axis('tight')
    plt.show()


# Plot the ROC (Receiver Operator Characteristic) Curves of the four classifiers
def graph_roc_curve(classifiers=None, classifier_names=None, path_data=None):
    import os
    from sklearn.externals import joblib  # For loading classifiers
    from scipy import interp  # For interpolation

    # If no classifiers specified, populate list with all four
    if classifiers is None:
        rbfsvm = joblib.load(os.path.join(os.path.sep, 'pythesis', 'classifiers', 'svm', 'rbf', 'net_speed', 'scaled', 'rbf_0.999929972393', 'rbf_0.999929972393.pkl'))
        nn = joblib.load(os.path.join(os.path.sep, 'pythesis', 'classifiers', 'nn', 'net_speed', 'scaled', '0.998504003751', '0.998504003751.pkl'))
        abc = joblib.load(os.path.join(os.path.sep, 'pythesis', 'classifiers', 'abc', 'net_speed', 'scaled', '0.982735815716', '0.982735815716.pkl'))
        rfc = joblib.load(os.path.join(os.path.sep, 'pythesis', 'classifiers', 'rfc', 'net_speed', 'scaled', '0.999961802902', '0.999961802902.pkl'))
        classifiers = [rbfsvm, nn, abc, rfc]

    # If no names specified, populate list with all four
    if classifier_names is None:
        classifier_names = ['rbfsvm', 'nn', 'abc', 'rfc']

    # If no path data specified, load the pickled data
    if path_data is None:
        curated_paths_name = os.path.join(os.path.sep, 'pythesis', 'curated_paths', 'all_paths_Jul13-Aug14.pkl')
        x_test, y_test = joblib.load(curated_paths_name)
    else:
        x_test, y_test = path_data[0], path_data[1]

    # Pre-process the data
    y_test = label_binarize(y_test, numpy.unique(y_test))
    n_classes = y_test.shape[1]

    # Scale the data
    scaler = joblib.load(os.path.join(os.path.sep, 'pythesis', 'classifiers', 'rfc', 'net_speed', 'scaled', '0.999961802902', 'scaler.pkl'))
    x_test = scaler.transform(x_test)

    # Enumerate through each classifier, and plot its ROC curve
    for i, clf in enumerate(classifiers):
        y_score = clf.predict_proba(x_test)
        fpr = dict()
        tpr = dict()
        roc_auc = dict()
        for n in range(n_classes):
            fpr[n], tpr[n], _ = roc_curve(y_test[:, n], y_score[:, n])
            roc_auc[n] = auc(fpr[n], tpr[n])

        # aggregate the false positives
        all_fpr = numpy.unique(numpy.concatenate([fpr[n] for n in range(n_classes)]))

        # interpolate all ROC curves
        mean_tpr = numpy.zeros_like(all_fpr)
        for n in range(n_classes):
            mean_tpr += interp(all_fpr, fpr[n], tpr[n])

        # Average and compute AUC
        mean_tpr /= n_classes

        fpr["macro"] = all_fpr
        tpr["macro"] = mean_tpr
        roc_auc["macro"] = auc(fpr["macro"], tpr["macro"])

        # Plot the ROC curve
        plt.plot(fpr["macro"], tpr["macro"], label='{0} (area = {1:0.2f})'
                                           ''.format(classifier_names[i], roc_auc["macro"]))

    # Add a diagonal, labels, a title, and a legend to the graph
    plt.plot([0, 1], [0, 1], 'k--')
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.05])
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.title('ROC Curves for Vessel Behavior')
    plt.legend(loc="lower right")
    plt.show()
