# Thesis Project - Machine Learning Behavior Categorization
# Author - Devin Frey

from sklearn import svm  # For SVM calculations
from sklearn import preprocessing  # For preprocessing SVM data
from sklearn.model_selection import StratifiedKFold
from sklearn.model_selection import cross_val_score
from sklearn.model_selection import GridSearchCV
from sklearn.externals import joblib  # For pickling classifiers

from sklearn.ensemble import RandomForestClassifier
from sklearn.ensemble import AdaBoostClassifier
from sklearn.neural_network import MLPClassifier

import os.path  # For creating filepaths

from datetime import datetime  # For manipulating datetime objects
import psycopg2  # Postgre driver

from time import time  # For timing of functions

from PlottingFunctions import *  # For plotting classifier data on a map
from GraphingFunctions import *  # For graphing classifier data on a grid

# String to connect to VMS Postgres database
conn_string = "host='localhost' dbname='vms' user='postgres' password='password'"


# Test the accuracy of the four classifiers (SVM, RFC, AdaBoost, and NN), and pickle them
def run_testing_suite(grid_searching=True, pickling=False):
    # Load the training/testing data
    curated_paths_name = os.path.join(os.path.sep, 'pythesis', 'curated_paths', 'all_paths_Jul13-Aug14.pkl')
    speeds, behaviors = joblib.load(curated_paths_name)

    # Create the default scaler, then scale the speeds between -1 and 1
    scaler = preprocessing.MinMaxScaler((-1, 1))
    speeds = scaler.fit_transform(speeds, behaviors)

    print("NN")
    # Neural Network
    if grid_searching:
        nn = MLPClassifier(solver='lbfgs', random_state=144)
        param_grid = {'alpha': [0.0001, 0.001, 0.1, 1, 10, 100, 1000], 'hidden_layer_sizes': [1, 10, 100, 1000]}
        optimal_params = perform_grid_search(param_grid, nn, speeds, behaviors)
        nn = MLPClassifier(solver='lbfgs', random_state=144, alpha=optimal_params['alpha'], hidden_layer_sizes=optimal_params['hidden_layer_sizes'])
    else:
        nn = MLPClassifier(solver='lbfgs', alpha=0.0001, hidden_layer_sizes=100, random_state=144)

    # Pickle the classifier
    if pickling:
        score = str(cross_val_score(nn, speeds, behaviors, cv=10, n_jobs=-1).mean())
        print("NN SCORE = " + score)
        nn = nn.fit(speeds, behaviors)
        classifier_directory = os.path.join(os.path.sep, 'pythesis', 'classifiers', 'nn', 'net_speed', 'scaled' if scaler else 'unscaled', score)
        if not os.path.exists(classifier_directory):
            os.makedirs(classifier_directory)
        fp = os.path.join(classifier_directory, score + '.pkl')
        joblib.dump(nn, fp)
        # save the scaler too, if it exists
        if scaler:
            fp = os.path.join(classifier_directory, 'scaler.pkl')
            joblib.dump(scaler, fp)

    print("RFC")
    # Random Forest Classifier
    if grid_searching:
        rfc = RandomForestClassifier(random_state=144)
        param_grid = {'n_estimators': [1, 10, 100, 1000], 'min_samples_leaf': [1, 10, 100, 1000]}
        optimal_params = perform_grid_search(param_grid, rfc, speeds, behaviors)
        rfc = RandomForestClassifier(n_estimators=optimal_params['n_estimators'], min_samples_leaf=optimal_params['min_samples_leaf'], random_state=144)
    else:
        rfc = RandomForestClassifier(n_estimators=1000, min_samples_leaf=1, random_state=144)

    # Pickle the classifier
    if pickling:
        score = str(cross_val_score(rfc, speeds, behaviors, cv=10, n_jobs=-1).mean())
        print("RFC SCORE = " + score)
        rfc = rfc.fit(speeds, behaviors)
        classifier_directory = os.path.join(os.path.sep, 'pythesis', 'classifiers', 'rfc', 'net_speed', 'scaled' if scaler else 'unscaled', score)
        if not os.path.exists(classifier_directory):
            os.makedirs(classifier_directory)
        fp = os.path.join(classifier_directory, score + '.pkl')
        joblib.dump(rfc, fp)
        # save the scaler too, if it exists
        if scaler:
            fp = os.path.join(classifier_directory, 'scaler.pkl')
            joblib.dump(scaler, fp)

    print("SVM")
    # Support Vector Machine, with a Radial Basis Function kernel
    if grid_searching:
        rbfsvm = svm.SVC(kernel='rbf', probability=True, random_state=144)
        param_grid = {'C': [1, 10, 100, 1000, 10000, 100000], 'gamma': [0.0001, 0.001, 0.1, 1, 10, 100, 1000]}
        optimal_params = perform_grid_search(param_grid, rbfsvm, speeds, behaviors)
        rbfsvm = svm.SVC(kernel='rbf', probability=True, C=optimal_params['C'], gamma=optimal_params['gamma'], random_state=144)
    else:
        rbfsvm = svm.SVC(kernel='rbf', C=100000, gamma=1000, probability=True, random_state=144)

    # Pickle the classifier
    if pickling:
        score = str(cross_val_score(rbfsvm, speeds, behaviors, cv=10, n_jobs=-1).mean())
        print("SVM SCORE = " + score)
        rbfsvm = rbfsvm.fit(speeds, behaviors)
        classifier_directory = os.path.join(os.path.sep, 'pythesis', 'classifiers', 'svm', 'rbf', 'net_speed', 'scaled' if scaler else 'unscaled', 'rbf' + '_' + score)
        if not os.path.exists(classifier_directory):
            os.makedirs(classifier_directory)
        fp = os.path.join(classifier_directory, 'rbf_' + str(score) + '.pkl')
        joblib.dump(rbfsvm, fp)
        # save the scaler too, if it exists
        if scaler:
            fp = os.path.join(classifier_directory, 'scaler.pkl')
            joblib.dump(scaler, fp)

    print("ADABOOST")
    # AdaBoost
    if grid_searching:
        abc = AdaBoostClassifier(random_state=144)
        param_grid = {'n_estimators': [1, 10, 25, 50, 100], 'learning_rate': [0.0001, 0.001, 0.1, 1]}
        optimal_params = perform_grid_search(param_grid, abc, speeds, behaviors)
        abc = AdaBoostClassifier(n_estimators=optimal_params['n_estimators'], learning_rate=optimal_params['learning_rate'], random_state=144)
    else:
        abc = AdaBoostClassifier(n_estimators=25, learning_rate=1, random_state=144)

    # Pickle the classifier
    if pickling:
        score = str(cross_val_score(abc, speeds, behaviors, cv=10, n_jobs=-1).mean())
        print("ABC SCORE = " + score)
        abc = abc.fit(speeds, behaviors)
        classifier_directory = os.path.join(os.path.sep, 'pythesis', 'classifiers', 'abc', 'net_speed', 'scaled' if scaler else 'unscaled', score)
        if not os.path.exists(classifier_directory):
            os.makedirs(classifier_directory)
        fp = os.path.join(classifier_directory, score + '.pkl')
        joblib.dump(abc, fp)
        # save the scaler too, if it exists
        if scaler:
            fp = os.path.join(classifier_directory, 'scaler.pkl')
            joblib.dump(scaler, fp)

    input("Testing run complete. Press enter to continue...")


# Method to test the trained and scored SVM against points with the "UNKNOWN" behavior type
def plot_against_unknown(classifier=None, cursor=None, confidence_threshold=None, scaler=None):
    if not classifier:
        classifier = joblib.load(os.path.join(os.path.sep, 'pythesis', 'classifiers', 'rfc', 'net_speed', 'scaled', '0.999961802902', '0.999961802902.pkl'))
    if not cursor:
        conn = psycopg2.connect(conn_string)
        cursor = conn.cursor()
    if not confidence_threshold:
        confidence_threshold = 0.5
    # Query for the speeds of points with unknown behavior
    unknown_query_string = "SELECT speed, net_speed FROM vessel_point WHERE vessel_behavior_id = 0"
    cursor.execute(unknown_query_string)
    unknown_points = numpy.ravel(cursor.fetchall())
    unknown_points = unknown_points.reshape(int(unknown_points.size/2), 2)
    # If a scaler is being used, scale the points
    if scaler:
        unknown_points = scaler.transform(unknown_points)
    # Use the classifier to predict the behavior of the unknown points
    probable_behaviors = classifier.predict_proba(unknown_points)
    predicted_behaviors = numpy.zeros(probable_behaviors.shape[0])
    for index, value in enumerate(probable_behaviors):
        # If the most probable behavior does not meet the confidence threshold, set it to unknown
        if max(probable_behaviors[index]) < confidence_threshold:
            predicted_behaviors[index] = 0
        # Else, set the behavior at this index to the likeliest behavior
        else:
            predicted_behaviors[index] = numpy.argmax(probable_behaviors[index]) + 1
    # plot_classifier_predictions(svm, unknown_points, predicted_behaviors, column)
    # plot_predicted_behavior(cursor, predicted_behaviors)


# Method to retrieve speed and behavior data from the database
def gather_parameters(path_ids=None, cursor=None):
    if not cursor:
        conn = psycopg2.connect(conn_string)
        cursor = conn.cursor()

    # If path IDs are given, query for the speeds of points in those paths
    if path_ids:
        path_query_string = "SELECT speed, net_speed FROM vessel_point WHERE vessel_path_id IN " + str(path_ids)
        cursor.execute(path_query_string)
        speeds = numpy.ravel(cursor.fetchall())
        speeds = speeds.reshape(int(speeds.size/2), 2)

        behavior_query_string = "SELECT vessel_behavior_id FROM vessel_point WHERE vessel_path_id IN " + str(path_ids)
        cursor.execute(behavior_query_string)
        behaviors = numpy.ravel(cursor.fetchall())
    # If no path IDs are given, query for all paths without unknown points
    else:
        unknown_paths = "(SELECT DISTINCT vessel_path_id FROM vessel_point WHERE vessel_behavior_id = 0)"
        path_query_string = "SELECT speed, net_speed FROM vessel_point WHERE vessel_path_id NOT IN " + unknown_paths
        cursor.execute(path_query_string)
        speeds = numpy.ravel(cursor.fetchmany(1000))
        speeds = speeds.reshape(int(speeds.size/2), 2)
        behavior_query_string = "SELECT vessel_behavior_id FROM vessel_point WHERE vessel_path_id NOT IN " + unknown_paths
        cursor.execute(behavior_query_string)
        behaviors = numpy.ravel(cursor.fetchmany(1000))

    return speeds, behaviors


# Perform a grid search and return the best parameters
def perform_grid_search(param_grid, classifier, paths, behaviors):
    # Record the starting time
    start_time = time()
    # Use 3-fold cross validation
    skf = StratifiedKFold(n_splits=3)
    # Create the grid, and fit the data to it
    grid = GridSearchCV(classifier, param_grid=param_grid, cv=skf, n_jobs=1)
    grid.fit(paths, behaviors)
    grid_search_time = (time() - start_time) / 60
    # Print out the results and execution time
    write_output("Classifier = %s, Time Searching = %s, Best Params = %s, Best Score = %s" % (type(classifier).__name__, grid_search_time, grid.best_params_, grid.best_score_))
    return grid.best_params_


# Method to write output of grid search to a file
def write_output(output_text):
    output_file = open('grid_search_output.txt', 'a')
    # Record the date and time of the output
    output_file.write('\n' + str(datetime.now()) + '\n')
    output_file.write(output_text + '\n')
    output_file.close()
