# Thesis Project - VMS Console
# File: Path.py
#
# Description: For each vessel, a path object can be created which
# contains a list of trips, which are logically clustered GPS points
# that convey a travel route over closely timed points.
#
# Author: David Gallegos
# Last Updated: 08/30/2014

import numpy         #Provides mean
from Point import * #Supporting class that contains the fields of a GPS record.


class Path:
    MAX_SEC_BETWEEN_POINTS = 1200 #maximum seconds between two points to be considered part of the same trip
    IGNORE_NON_GULF_OF_MEXICO = True
    MIN_CONSIDERED_TRIP_SIZE = 10 #minimum number of points in a trip to be considered as a valid trip.
    MAX_VALID_SPEED = 50          #maximum speed in m/s that a vessel can actually conceivably do, to filter glitches.
    
    def __init__(self):
        self.raw_data = set() #prevent exact duplicates - Point class defines equality criteria
        self.trips = []
        self.calculated = True

    def add_data(self, data):
        '''Adds a data point for this vessel'''
        if not isinstance(data, Point):
            print("Error: Path only accepts objects of type Point.")
            return
        
        #check very rough square of gulf of mexico
        if Path.IGNORE_NON_GULF_OF_MEXICO and (data.latitude > 32 or data.latitude < 22 or data.longitude > -80 or data.longitude < -98):
            #not in gulf of mexico - ignore
            pass
        else:
            self.raw_data.add(data)
            self.calculated = False

    def get_trips(self):
        '''Returns a list of trips for this vessel, does speed & behavior calculations if needed.'''
        if not (self.calculated):
            self.calculate()
        return self.trips

    def calculate(self):
        '''Calculates each point's speed and behavior by comparing it to the item previous.'''
        del self.trips[:]
        
        sorted_data = sorted(self.raw_data)

        #Arrange data into trips, with several considerations:
        #   1) If the points are sufficiently far apart, consider it a new trip
        #   2) If the timestamp is the same, discard it as a duplicate
        for p in sorted_data:
            if len(self.trips) == 0:
                self.trips.append([p])
            else:
                diff_seconds = (p.timestamp - self.trips[-1][-1].timestamp).total_seconds()
                assert diff_seconds >= 0, "Diff_seconds is negative" #should never be negative

                if diff_seconds > Path.MAX_SEC_BETWEEN_POINTS:
                    #Treat this as a new trip, too many seconds have passed between points.
                    self.trips.append([p])
                else:
                    #discard if the timestamp is identical. Note that we're already isolated to one vessel.
                    if diff_seconds >= 1:
                        #Calculate the point's speed compared to the last point in the trip.
                        #There will always be a previous point at this step.
                        p.calculate_speed(self.trips[-1][-1])
                        
                        #Do not include points with speed over an impossible number, this is glitched data.
                        if p.speed < Path.MAX_VALID_SPEED:
                            self.trips[-1].append(p) #Add this point to the end
        
        #sweep out anything with less points than the minimum. Does not effect raw data, only the trips info
        self.trips = [t for t in self.trips if len(t) >= Path.MIN_CONSIDERED_TRIP_SIZE]

        for t in self.trips:
            #assume 1st data point was doing same thing as 2nd data point
            t[0].speed, t[0].duration = t[1].speed, t[1].duration

            #calculate a sliding window of net distance covered by vessel on this trip
            window_start = 0
            window_end = Path.MIN_CONSIDERED_TRIP_SIZE - 1
            while (window_end < len(t)):
                #Total distance covered from window beginning to end
                delta_dist = t[window_start].great_circles_distance(t[window_end])
                
                #Total time elapsed from beginning to end
                delta_time = (t[window_end].timestamp - t[window_start].timestamp).total_seconds()
                
                #net_speed, that is the speed over the window's time
                net_speed = delta_dist / delta_time
                
                for i in range(window_start, window_end+1):
                    #For each point, append this net speed to its list which will be averaged later.
                    t[i].net_speeds.append(net_speed)
                window_start += 1
                window_end += 1

            for idx, pt in enumerate(t):
                #Get the average of all points generated by the net speeds in the sliding window
                pt.net_speed = numpy.mean(pt.net_speeds)
                
                #Determine what the vessel was doing based on aggregate data
                pt.calculate_behavior()
                
                #Get the movement angles of the points
                #If there is no previous, skip this iteration
                if idx - 1 < 0:
                    continue
                else:
                    prev = t[idx - 1]
                #If there is no next point, skip this iteration
                if idx + 1 == len(t):
                    continue
                else:
                    next = t[idx + 1]
                pt.calculate_movement_angle(prev, next)

            # Once all other points have calculated their movement angles,
            # set the movement angles of the first point to the second point,
            # and set the movement angle of the last point to the second to last point
            t[0].movement_angle = t[1].movement_angle
            t[-1].movement_angle = t[-2].movement_angle
        self.calculated = True