# Thesis Project - SVM Behavior Categorization
# Author - Devin Frey

import psycopg2  # Postgre driver
from mpl_toolkits.basemap import Basemap
from sklearn.externals import joblib  # For pickling classifiers
import os.path  # For creating filepaths
import matplotlib.pyplot as plt  # For plotting data
import numpy  # For ravel function
import sys  # For printing errors
import urllib
from collections import OrderedDict  # For fixing the map legend

from PIL import Image
from io import BytesIO

# Configure to target VMS Postgres database
conn_string = "host='localhost' dbname='vms' user='postgres' password='password'"


# Method to plot a map of the coastal waters of Louisiana, showing the locations of vessel activity and the behavior
# the deterministic classifier predicts the vessel was exhibiting at that time
def plot_gulf_map(path_ids, cursor):
    # Create the map, drawing the coastlines, continents, and other map features
    m = Basemap(projection='cyl', llcrnrlat=28.5, urcrnrlat=30.5, llcrnrlon=-92.2, urcrnrlon=-89, resolution='f')
    m.drawcoastlines()
    m.fillcontinents(color='coral', lake_color='aqua')
    m.drawparallels(numpy.arange(22, 32.1, 0.2), labels=[False, True, True, False])
    m.drawmeridians(numpy.arange(-98, -79.9, 0.2), labels=[True, False, False, True])
    m.drawmapboundary(fill_color='aqua')
    plt.title("Vessel Behavior in Louisiana Waters")
    # Convert path ids to a string
    path_ids = str(tuple(path_ids))
    # Markers for the five behaviors
    map_markers = ['bo', 'co', 'yo', 'mo', 'ro']
    # For each behavior, query for the coordinates of points exhibiting that behavior,
    # then plot them on the map
    for behavior_id in range(0, 5):
        window_query_string = "SELECT ST_X(coords), ST_Y(coords) FROM vessel_point " \
                              "WHERE vessel_behavior_id = " + str(behavior_id) + " AND vessel_path_id IN " + path_ids
        cursor.execute(window_query_string)
        coordinates = numpy.ravel(cursor.fetchall())
        lons = []
        lats = []
        # All lats will be positive, and all lons will be negative
        for value in coordinates:
            if value < 0:
                lons.append(value)
            else:
                lats.append(value)
        lons, lats = m(lons, lats)
        print("Plotting behavior " + str(behavior_id))
        m.plot(lons, lats, map_markers[behavior_id])
    plt.show()


# Method to plot each vessel's behavior in public oyster grounds, using a non-deterministic classifier
def plot_each_vessel(cursor=None):
    if cursor is None:
        conn = psycopg2.connect(conn_string)
        cursor = conn.cursor()

    vessel_id_query_string = "SELECT DISTINCT vessel_id FROM vessel_path ORDER BY vessel_id"
    cursor.execute(vessel_id_query_string)
    vessel_ids = cursor.fetchall()
    vessel_ids = [i[0] for i in vessel_ids]
    for vessel_id in vessel_ids:
        plot_single_vessel(cursor, vessel_id, False)


# Plot each vessel's paths with unknown points, using a non-deterministic classifier
def plot_each_unknown_vessel(cursor=None):
    if cursor is None:
        conn = psycopg2.connect(conn_string)
        cursor = conn.cursor()

    vessel_id_query_string = "SELECT DISTINCT vessel_id, count(vessel_id) FROM vessel_path WHERE vessel_path_id IN (SELECT vessel_path_id FROM vessel_point WHERE vessel_behavior_id = 0) GROUP BY vessel_id ORDER BY count(vessel_id) DESC"
    cursor.execute(vessel_id_query_string)
    vessel_ids = cursor.fetchall()
    vessel_ids = [i[0] for i in vessel_ids]
    for vessel_id in vessel_ids:
        plot_single_vessel(cursor, vessel_id, True)


# Plot each vessel path in order from largest to smallest
def plot_each_path(cursor=None, starting_path_id=None):
    # Begin from where the program left off
    starting_path_id = 167
    if cursor is None:
        conn = psycopg2.connect(conn_string)
        cursor = conn.cursor()

    path_id_query_string = "SELECT vessel_path_id, count(vessel_point_id) FROM vessel_point GROUP BY vessel_path_id ORDER BY count(vessel_point_id) DESC"
    cursor.execute(path_id_query_string)
    path_ids = cursor.fetchall()
    path_ids = [i[0] for i in path_ids]
    # If a starting path id is given, slice off all ids occurring before the given id
    if starting_path_id:
        path_ids = path_ids[path_ids.index(starting_path_id):]
    # Make a map for each vessel path
    for path_id in path_ids:
        plot_single_path(cursor, path_id, False)


# Method to plot a single vessel's behavior in public oyster grounds
def plot_single_vessel(cursor=None, vessel_id=None, unknown_only=False):
    if cursor is None:
        conn = psycopg2.connect(conn_string)
        cursor = conn.cursor()

    if vessel_id is None:
        vessel_id = input('Please enter a vessel id:')
        # Cast to int for sanitization purposes
        vessel_id = int(vessel_id)

    # If only plotting unknowns, append the following to the behavior_query_string
    if unknown_only:
        addendum = " AND vessel_behavior_id = 0"
    else:
        addendum = ""

    print("MAPPING PATH FOR VESSEL " + str(vessel_id))
    # Query for each path that belongs to this vessel
    vessel_path_query_string = "SELECT vessel_path_id FROM vessel_path WHERE vessel_id = " + str(vessel_id)
    cursor.execute(vessel_path_query_string)
    vessel_path_ids = cursor.fetchall()
    vessel_path_ids = tuple([i[0] for i in vessel_path_ids])
    # If the vessel has only one path, adjust the string
    if len(vessel_path_ids) < 2:
        vessel_path_ids = str("(" + str(vessel_path_ids[0]) + ")")
    # Query for the coordinates of each point in this vessel's paths
    coords_query_string = "SELECT ST_X(coords), ST_Y(coords) FROM vessel_point WHERE vessel_path_id IN " + str(vessel_path_ids)
    cursor.execute(coords_query_string)
    coords = cursor.fetchall()
    lons = []
    lats = []
    for values in coords:
        lons.append(values[0])
        lats.append(values[1])
    # The difference between the maximum and minimum latitude values
    lats_diff = max(lats) - min(lats)
    # The difference between the maximum and minimum longitude values
    lons_diff = max(lons) - min(lons)

    map_markers = ['bo', 'co', 'yo', 'mo', 'ro']

    # Create the map. The boundaries will be slightly greater than the max/min coordinates.
    m = Basemap(projection='cyl', llcrnrlat=min(lats) - lats_diff/5, urcrnrlat=max(lats) + lats_diff/5, area_thresh=0.1,
                llcrnrlon=min(lons) - lons_diff/5, urcrnrlon=max(lons) + lons_diff/5, resolution='f', epsg=2801)
    m.arcgisimage(service='World_Imagery', xpixels=1500000, dpi=1500000)
    mlons, mlats = m(lons, lats)
    # Gather the speeds of the points in the paths
    parameters_query_string = "SELECT speed, net_speed FROM vessel_point WHERE vessel_path_id IN " + str(vessel_path_ids) + " AND vessel_behavior_id != 0"
    cursor.execute(parameters_query_string)
    parameters = numpy.ravel(cursor.fetchall())
    parameters = parameters.reshape(int(parameters.size/2), 2)
    # Load the classifier, then predict the behaviors of the vessel's points
    clf = joblib.load(os.path.join(os.path.sep, 'pythesis', 'classifiers', 'rbf', 'net_speed', 'unscaled', 'rbf_0.999696125748', 'rbf_0.999696125748.pkl'))
    clf_predictions = clf.predict(parameters)
    # Smooth the predictions
    clf_predictions = smooth(clf_predictions)
    for index, value in enumerate(clf_predictions):
        m.plot(mlons[index], mlats[index], map_markers[int(value)])
    plt.show()


# Method to plot a single vessel's behavior in public oyster grounds
def plot_single_path(cursor=None, path_id=None, unknown_only=False):
    # clear the figure, so old figures do not "bleed" into new ones
    plt.clf()
    # If a cursor was not given, instantiate a new one
    if cursor is None:
        conn = psycopg2.connect(conn_string)
        cursor = conn.cursor()

    # If no path id was given, query the user for one
    if path_id is None:
        path_id = input('Please enter a path id:')
        # Cast to int for sanitization purposes
        path_id = int(path_id)

    # If only plotting unknowns, append the following to the behavior_query_string
    if unknown_only:
        addendum = " AND vessel_behavior_id = 0"
    else:
        addendum = ""

    print("MAPPING PATH " + str(path_id))
    # Query for the coordinates of each point in the path
    coords_query_string = "SELECT ST_X(coords), ST_Y(coords) FROM vessel_point WHERE vessel_path_id = " + str(path_id)
    cursor.execute(coords_query_string)
    coords = cursor.fetchall()
    lons = []
    lats = []
    # Gather the lat and lon coordinates
    for values in coords:
        lons.append(values[0])
        lats.append(values[1])
    # The difference between the maximum and minimum latitude values
    lats_diff = max(lats) - min(lats)
    # The difference between the maximum and minimum longitude values
    lons_diff = max(lons) - min(lons)
    # The maximum and minimum boundaries of the map
    map_max_lat = max(lats) + lats_diff/8
    map_min_lat = min(lats) - lats_diff/8
    map_max_lon = max(lons) + lons_diff/8
    map_min_lon = min(lons) - lons_diff/8
    # Markers for each vessel behavior
    map_markers = ['bo', 'co', 'yo', 'mo', 'ro']
    # Make the figure large enough that points do not overlap
    plt.figure(figsize=(21.00, 11.92))
    # Generate the map and its boundaries, then draw its parallels and meridians
    m = Basemap(projection='cyl', llcrnrlat=map_min_lat, urcrnrlat=map_max_lat, area_thresh=0.1,
                llcrnrlon=map_min_lon, urcrnrlon=map_max_lon, resolution='f', epsg=2801)
    # Draw the map's parallels and meridians
    parallels = numpy.arange(min(lats), map_max_lat, lats_diff/2.0)
    m.drawparallels(parallels, labels=[False, True, True, False])
    meridians = numpy.arange(min(lons), map_max_lon, lons_diff/2.0)
    m.drawmeridians(meridians, labels=[True, False, False, True])

    # Attempt to pull tiles from arcgis. If an http error occurs, skip to next path
    try:
        m.arcgisimage(service='World_Imagery', xpixels=1500000, dpi=1500000)
    except urllib.error.HTTPError:
        print("HTTPError occurred, skipping path " + str(path_id))
        return

    # convert the lons and lats to basemap lat and lon values
    mlons, mlats = m(lons, lats)
    # Query for the path's speeds
    parameters_query_string = "SELECT speed, net_speed FROM vessel_point WHERE vessel_path_id = " + str(path_id) + " AND vessel_behavior_id != 0"
    cursor.execute(parameters_query_string)
    parameters = numpy.ravel(cursor.fetchall())
    # Reshape the array so the x value denotes which point and the y value denotes either speed or net speed
    parameters = parameters.reshape(int(parameters.size/2), 2)
    # Load the optimal classifier, and its scaler
    clf = joblib.load(os.path.join(os.path.sep, 'pythesis', 'classifiers', 'rfc', 'net_speed', 'scaled', '0.999961802902', '0.999961802902.pkl'))
    scaler = joblib.load(os.path.join(os.path.sep, 'pythesis', 'classifiers', 'rfc', 'net_speed', 'scaled', '0.999961802902', 'scaler.pkl'))
    # Transform the speeds to fit the scaler, then predict the behaviors of points based on their speeds
    parameters = scaler.transform(parameters)
    clf_predictions = clf.predict(parameters)
    # Smooth the predictions
    clf_predictions = smooth(clf_predictions)
    # Plot and label the path's points
    labels = ["Unknown", "Docked", "Fishing", "Traveling", "Poling"]
    for index, value in enumerate(clf_predictions):
        m.plot(mlons[index], mlats[index], map_markers[int(value)], label=labels[int(value)])
    #print("LATs = " + str(min(lats)) + "/" + str(max(lats)) + " / LONs = " + str(min(lons)) + "/" + str(max(lons)))
    # The legend will have a label for each point; the lines below remove duplicates from the legend
    handles, labels = plt.gca().get_legend_handles_labels()
    by_label = OrderedDict(zip(labels, handles))
    plt.legend(by_label.values(), by_label.keys(), loc="lower right", prop={'size': 12})
    # Save png file
    plt.savefig('C:/pythesis/figures/vessel path plots/pngs/' + str(path_id) + '.png')
    # Save the image to memory
    png = BytesIO()
    plt.savefig(png, format='png')
    # Convert the image to TIFF
    tiff = Image.open(png)
    tiff.save('C:/pythesis/figures/vessel path plots/tiffs/' + str(path_id) + '.tiff')
    plt.close()
    png.close()
    tiff.close()


# Plot every point where a specific behavior was being exhibited
def plot_behavior(cursor, behavior_id):
    # Query for the coordinates of each point exhibiting the selected behavior
    point_query_string = "SELECT ST_X(coords), ST_Y(coords) FROM vessel_point WHERE vessel_behavior_id = " + str(behavior_id)
    cursor.execute(point_query_string)
    coords = cursor.fetchall()
    lons = []
    lats = []
    # Gather the lats and lons of the points
    for values in coords:
        lons.append(values[0])
        lats.append(values[1])
    # Draw the map. The boundaries of the map will be 1/5 greater than the difference between the max and min lat and the max and min lon
    m = Basemap(projection='cyl', llcrnrlat=min(lats) - max(lats)/min(lats)/5, urcrnrlat=max(lats) + max(lats)/min(lats)/5, llcrnrlon=min(lons) - max(lons)/min(lons)/5, urcrnrlon=max(lons) + max(lons)/min(lons)/5, resolution='f')
    lons, lats = m(lons, lats)
    # Markers for the five behaviors. Only the one corresponding to behavior_id will be used
    map_markers = ['bo', 'co', 'yo', 'mo', 'ro']
    # Plot the points of the chosen behavior
    m.plot(lons, lats, map_markers[behavior_id])
    # Draw the details of the map
    m.drawcoastlines()
    m.fillcontinents(color='coral', lake_color='aqua')
    m.drawmapboundary(fill_color='aqua')
    # Title the map based on the name of the behavior_id given, then show the map
    behaviors = ["Unknown", "Docked", "Fishing", "Traveling", "Poling"]
    plt.title(behaviors[behavior_id] + " Points, Jul13-Aug14")
    plt.show()


# Plot the SVM's predictions of unknown point behaviors
def plot_predicted_behavior(cursor, predicted_behaviors):
    # Query for the coordinates of points with unknown behavior
    unknown_behavior_id = "0"
    point_query_string = "SELECT ST_X(coords), ST_Y(coords) FROM vessel_point WHERE vessel_behavior_id = " + unknown_behavior_id
    cursor.execute(point_query_string)
    coords = cursor.fetchall()
    lons = []
    lats = []
    for values in coords:
        lons.append(values[0])
        lats.append(values[1])
    # Draw the map. The boundaries of the map will be 1/5 greater than the difference between the max and min lat and the max and min lon
    m = Basemap(projection='cyl', llcrnrlat=min(lats) - max(lats)/min(lats)/5, urcrnrlat=max(lats) + max(lats)/min(lats)/5, llcrnrlon=min(lons) - max(lons)/min(lons)/5, urcrnrlon=max(lons) + max(lons)/min(lons)/5, resolution='f')
    lons, lats = m(lons, lats)
    # Markers for the five behaviors
    map_markers = ['bo', 'co', 'yo', 'mo', 'ro']
    # Iterate through the map markers. Each marker's index corresponds to its behavior's ID in the database.
    for behavior_id, marker in enumerate(map_markers):
        # Find all behaviors corresponding to the current marker, then plot them on the map
        indices = list((index for index, value in enumerate(predicted_behaviors) if value == behavior_id))
        m.plot([lons[index] for index in indices], [lats[index] for index in indices], map_markers[behavior_id])
    # Add details to the map, then show it
    m.drawcoastlines()
    m.fillcontinents(color='coral', lake_color='aqua')
    m.drawmapboundary(fill_color='aqua')
    plt.title("Predicted Behavior of Unknown Points, Jul13-Aug14")
    plt.show()


# Helper method to smooth an SVM's predictions of a vessel's behavior
def smooth(predictions):
    for index, prediction in enumerate(predictions):
        # If this is the first or last element, skip this iteration
        if index == 0 or index == len(predictions) - 1:
            continue
        # If the previous and next prediction are the same,
        if predictions[index - 1] == predictions[index + 1]:
            # If the current prediction differs from the previous,
            if prediction != predictions[index - 1]:
                # Set this behavior to the previous behavior
                predictions[index] = predictions[index - 1]
    return predictions

