# Thesis Project - VMS Console
# File: Point.py
#
# Description: A container for GPS points including position and time.
# Also provides methods to determine speed of travel and distance between
# itself and another point. It also can return a string behavior descriptor.
#
# Author: David Gallegos
# Last Updated: 08/30/2014

import math #Provides various mathematical functions

class Point:
    #Behavior constants
    UNKNOWN = 0
    DOCKED = 1
    FISHING = 2
    TRAVELING = 3
    POLING = 4
    
    def __init__(self, timestamp, latitude, longitude):
        self.timestamp = timestamp
        self.latitude = latitude
        self.longitude = longitude
        
        #below should not be part of eq or hash, it's a calculation, not an inherent property
        self.distance = 0 #in meters
        self.speed = 0
        self.movement_angle = 0
        self.behavior = Point.UNKNOWN
        self.duration = 0 #in s
        self.net_speeds = []
        self.net_speed = 0
        self.calculated = False

    #The following methods provide comparability between Points for sorting, hashing and direct comparison.
    def __hash__(self):
        return hash( (self.timestamp, self.latitude, self.longitude) )
    def __eq__(self, other):
        return isinstance(other, self.__class__) and (self.timestamp - other.timestamp).total_seconds() == 0
    def __ne__(self, other):
        return (not isinstance(other, self.__class__)) or (self.timestamp - other.timestamp).total_seconds() != 0        
    def __lt__(self, other):
        return isinstance(other, self.__class__) and (self.timestamp - other.timestamp).total_seconds() < 0
    def __le__(self, other):
        return isinstance(other, self.__class__) and (self.timestamp - other.timestamp).total_seconds() <= 0
    def __gt__(self, other):
        return isinstance(other, self.__class__) and (self.timestamp - other.timestamp).total_seconds() > 0
    def __ge__(self, other):
        return isinstance(other, self.__class__) and (self.timestamp - other.timestamp).total_seconds() >= 0

    def get_behavior(self):
        '''Returns string behavior descriptor based on point's attributes.'''
        if self.behavior == Point.DOCKED:
            return "Docked"
        elif self.behavior == Point.FISHING:
            return "Fishing"
        elif self.behavior == Point.TRAVELING:
            return "Traveling"
        elif self.behavior == Point.POLING:
            return "Poling"
        else:
            return "Unknown"
    
    def calculate_speed(self, other):
        '''Sets own's speed based on comparison with another point (The one prior to it).'''
        if not isinstance(other, self.__class__):
            raise TypeError("Must be comparing two Point objects to calculate behavior.")
        
        self.distance = self.great_circles_distance(other)
        self.duration = (self.timestamp - other.timestamp).total_seconds()
        
        #Ensure positive duration, this is a "can't happen" situation.
        assert self.duration > 0, "Duration is <= 0, must be positive. Calculate_speed expects a point in past."
        
        self.speed = self.distance / self.duration #m/s
        self.calculated = True

    def calculate_movement_angle(self, previous, next):
        '''Calculates this point's angle based on the lat/long coordinates of it and its two adjacent points'''
        if previous is None:
            previous = self # If previous or next is unavailable, use self. Make a deep copy if this causes issues.
        if next is None:
            next = self
        if not isinstance(previous, self.__class__):
            raise TypeError("Previous must be a Point object to calculate angle.")
        if not isinstance(next, self.__class__):
            raise TypeError("Next must be a Point object to calculate angle.")

        # Calculate the length of each edge of the triangle
        A = math.sqrt(math.pow((self.longitude - next.longitude), 2) + math.pow((self.latitude - next.latitude), 2))
        B = math.sqrt(math.pow((previous.longitude - next.longitude), 2) + math.pow((previous.latitude - next.latitude), 2))
        C = math.sqrt(math.pow((previous.longitude - self.longitude), 2) + math.pow((previous.latitude - self.latitude), 2))

        # Calculate the angle of the vertex where this point is
        if A <= 0 or C <= 0:
            #TODO: This is causing the "floor" of points with angle == 0 in the svm plots
            angle = 0
        else:
            alpha = math.atan2(previous.latitude - self.latitude, previous.longitude - self.longitude)
            beta = math.atan2(next.latitude - self.latitude, next.longitude - self.longitude)
            angle = math.degrees(beta - alpha)
            if (angle < 0):
                angle += 360

        self.movement_angle = angle

    def calculate_behavior(self):
        '''Calculate vessel behavior based on speed and net speed, both of which must be set already.'''

        if self.speed < 0.1 and self.net_speed < 0.1:
            # ~.2 knots
            self.behavior = Point.DOCKED
        elif self.speed < 0.5 and self.net_speed <= self.speed + 0.125:
            # <~.2 knots, a bit higher than diagonal and below.
            self.behavior = Point.POLING
        elif self.speed >= 0.5 and self.speed <= 1.75 and self.net_speed <= self.speed - 0.125:
            # ~1-3.4 knots, slightly below the diagonal
            self.behavior = Point.FISHING
        elif (self.speed >= 0.5 and self.net_speed >= self.speed + 0.125) or (self.speed > 2 and self.net_speed >= self.speed - 2):
            # Catch everything slightly above the diagonal when speed is 0.5 and higher, minus very low net speed
            self.behavior = Point.TRAVELING
        else:
            #unknown, it's a grey zone we're not going to try to interpret. Includes bizarre data and marginal data.
            pass
        
    def great_circles_distance(self, other):
        '''Calculates distance based on Haversine formula for great circles, returns in meters.'''
        if not isinstance(other, self.__class__):
            raise TypeError("Must be comparing two Point objects to calculate great circles distance")

        lat1 = self.latitude
        long1 = self.longitude
        lat2 = other.latitude
        long2 = other.longitude
        
        dlat = math.radians(lat2-lat1)
        dlong = math.radians(long2-long1)
        
        a = math.sin(dlat/2) * math.sin(dlat/2) + math.cos(math.radians(lat1)) \
            * math.cos(math.radians(lat2)) * math.sin(dlong/2) * math.sin(dlong/2)
        c = math.atan2(math.sqrt(a), math.sqrt(1-a))
    
        m_distance = 2 * c * 6371000 #earth radius in m
    
        return m_distance