# Thesis Project - VMS Console
# File: VMSC.py
#
# Description: The VMSC interacts with a Postgres database to allow analysis 
# and import of VMS data and provides console and Excel output options. This is
# the primary console file.
#
# Author: David Gallegos

import csv		#Provides simple comma separated value (CSV) output

from DbFunctions import * #Supporting class containing DB functionality.
from MLFunctions import * #Supporting class containing SVM functionality
from PlottingFunctions import *  # For plotting SVM data
from GraphingFunctions import *  # For graphing SVM data

#TUI constants and options
OPTION_KEY               = 0
CALL_FUNCTION            = 1
PROMPT_ENTER_TO_CONTINUE = 2
DISPLAY_PROMPT           = 3
dashes = '-' * 40
main_options = []
report_options = []
NO_FUNCTION              = 0

def main():
    #Initialize selection menus - Cannot be done globally because functions weren't defined yet.
    init_menus()
    
    #Check for Excel file program parameters.
    for arg_num in range(1, len(sys.argv)):
            import_data_from_excel(sys.argv[arg_num])
            
    if len(sys.argv) > 1:
        input("Press return key to continue...")
    
    sel_input = ''
    while sel_input != 'X':
        display_options(
            [
                "\n  VMS Data Management Console\n  Thesis Project - David X. Gallegos\n  SVM and Graphing Functionality Implemented by Devin Frey",
                print_statistics
            ],
            main_options
        )
        sel_input = process_input(main_options)

def data_report():
    '''Display report options'''
    sel_input = ''
    while sel_input != 'R':
        display_options(
            "Available VMSC Reports",
            report_options
        )
        sel_input = process_input(report_options)

def plotting_functions():
    '''Display mapping functions'''
    sel_input = ''
    while sel_input != 'R':
        display_options(
            "Mapping Functions:",
            plotting_options
        )
        sel_input = process_input(plotting_options)
    
def init_menus():
    '''Initializes menu options.'''
    global main_options
    global report_options
    global plotting_options
    
    main_options = [
        ['I', import_data_from_excel, True,  '(I)mport Data'],
        ['D', delete_all_data,        True,  '(D)elete All Data'],
        ['R', data_report,            False, '(R)eport on Data'],
        ['S', run_testing_suite,      False, 'Run Testing (S)uite'],
        ['P', plotting_functions,     False, '(P)lotting Functions'],
        ['X', NO_FUNCTION,            False, 'E(x)it']
    ]

    report_options = [
        ['A', activity_report,        True,  '(A)ctivity Report'],
        ['F', fishing_effort_report,  True,  '(F)ishing Effort Report'],
        ['R', NO_FUNCTION,            False, '(R)eturn to Main Menu']
    ]

    plotting_options = [
        ['E', plot_each_vessel,       False, 'Plot (E)ach Vessel'],
        ['U', plot_each_unknown_vessel,     False, 'Plot (U)nknown Vessels'],
        ['S', plot_single_vessel,     False, 'Plot (S)ingle Vessel'],
        ['P', plot_each_path,         False, 'Plot Each Vessel (P)ath'],
        ['C', graph_roc_curve,         False, 'Plot ROC (C)urve'],
        #[‘L’, versus_line_graph, False, 'Plot Versus (L)ine Graph'],
        ['R', NO_FUNCTION,            False, '(R)eturn to Main Menu']
    ]
        
def print_statistics():
    '''Displays statistical information about current data.'''
    queries = [
                "SELECT COUNT(DISTINCT vessel_id) FROM vessel_path",
                "SELECT COUNT(*) FROM vessel_path",
                "SELECT COUNT(*) FROM vessel_point"
              ]
    results = tuple(db_fetch(queries, None, True))
    print(
'''  Number of vessels in system: %d
  Number of trips in system:   %d
  Number of points in system:  %d''' % results)
    
def display_options(sections, options, cls=True):
    '''Displays a title, body and various options.'''
    if cls:
        os.system('cls')

    #If sections was passed as a string
    if not isinstance(sections, list):
        sections = [sections]
        
    #For each section, print it with dashes afterwards. Allows more diverse output
    for section in sections:
        if hasattr(section, '__call__'):
            #Argument is a callable function
            section()
        else:
            #Argument is intended to be printed
            print(section)
        print(dashes)

    #Print out each option key & display prompt
    print('  Press key associated with option:')
    for option_settings in options:
        print('  %s: %s' % (option_settings[OPTION_KEY], option_settings[DISPLAY_PROMPT]))

    print(dashes)
    
def process_input(options):
    err = False
    sel_option = input("  Choose option: ").upper()
    sel_item = []

    try:
        sel_item = [k for k in options if k[OPTION_KEY] == sel_option][0]
        call_func = sel_item[CALL_FUNCTION]
        
        if call_func:
            #Call the function associated with that option
            call_func()
    except IndexError:
        print("Invalid selection.")
        err = True

    if sel_option != "" and (err or sel_item[PROMPT_ENTER_TO_CONTINUE]):
        input("Press return key to continue...")

    return sel_option

def activity_report():
    '''Provide report on actual activity of vessels including speed, behavior and duration.'''
    results = run_report_query(
        '''
        SELECT vessel_id, point_timestamp, vessel_behavior_name, ST_Y(coords) latitude, ST_X(coords) longitude, duration_in_seconds, speed, net_speed, movement_angle
        FROM vessel_path
        INNER JOIN vessel_point on vessel_path.vessel_path_id = vessel_point.vessel_path_id
        INNER JOIN vessel_behavior ON vessel_point.vessel_behavior_id = vessel_behavior.vessel_behavior_id
        ''',
        '''
        ORDER BY vessel_id, point_timestamp
        '''
    )
    
    if len(results) == 0:
        print("No results for that query.")
    else:
        export_data_to_csv(["Vessel_ID", "Timestamp", "Behavior", "Latitude", "Longitude", "Duration", "Speed", "Net_Speed", "Movement_Angle"], results)

def fishing_effort_report():
    '''Provide information about how much time the vessel spent on each activity per day.'''
    results = run_report_query(
        '''
SELECT * FROM
(
    SELECT
        vpo.vessel_id,
        (vpo.vessel_path_start - (5 * '1 hour'::INTERVAL))::date trip_day,
        COALESCE(sum(path_fishing_time_hrs), 0.0) day_fishing_time_hrs,
        COALESCE(sum(path_travel_time_hrs), 0.0) day_travel_time_hrs,
        COALESCE(sum(path_docked_time_hrs), 0.0) day_docked_time_hrs,
        COALESCE(sum(path_poling_time_hrs), 0.0) day_poling_time_hrs,
        COALESCE(sum(path_unknown_time_hrs), 0.0) day_unknown_time_hrs
    FROM vessel_path vpo
    LEFT JOIN (
        SELECT round(SUM(duration_in_seconds) / 3600.0,1) path_fishing_time_hrs, vessel_path_id
        FROM vessel_point
        WHERE vessel_behavior_id = 2
        GROUP BY vessel_path_id
    ) vp1 ON vpo.vessel_path_id = vp1.vessel_path_id
    LEFT JOIN (
        SELECT round(SUM(duration_in_seconds) / 3600.0,1) path_travel_time_hrs, vessel_path_id
        FROM vessel_point
        WHERE vessel_behavior_id = 3
        GROUP BY vessel_path_id
    ) vp2 ON vpo.vessel_path_id = vp2.vessel_path_id
    LEFT JOIN (
        SELECT round(SUM(duration_in_seconds) / 3600.0,1) path_docked_time_hrs, vessel_path_id
        FROM vessel_point
        WHERE vessel_behavior_id = 1
        GROUP BY vessel_path_id
    ) vp3 ON vpo.vessel_path_id = vp3.vessel_path_id
    LEFT JOIN (
        SELECT round(SUM(duration_in_seconds) / 3600.0,1) path_poling_time_hrs, vessel_path_id
        FROM vessel_point
        WHERE vessel_behavior_id = 4
        GROUP BY vessel_path_id
    ) vp4 ON vpo.vessel_path_id = vp4.vessel_path_id
    LEFT JOIN (
        SELECT round(SUM(duration_in_seconds) / 3600.0,1) path_unknown_time_hrs, vessel_path_id
        FROM vessel_point
        WHERE vessel_behavior_id = 0
        GROUP BY vessel_path_id
    ) vp5 ON vpo.vessel_path_id = vp5.vessel_path_id
    GROUP BY vessel_id, (vpo.vessel_path_start - (5 * '1 hour'::INTERVAL))::date
    ORDER BY vessel_id, trip_day
) t
        ''',
        '',
        "trip_day"
    )
    if len(results) == 0:
        print("No results for that query.")
    else:
        export_data_to_csv(["Vessel_ID", "Trip_Date", "Fishing_Hrs", "Travel_Hrs", "Docked_Hrs", "Poling_Hrs", "Unknown_Hrs"], results)

def export_data_to_csv(title_row, records):
    '''Takes a title list and record list, writes to a prompted CSV file.'''
    output_file = input('Output file path/name: ')
    
    if output_file == "":
        output_file = "report"
    
    if str.lower(output_file)[-4:] != '.csv':
        output_file += '.csv'
    
    with open(output_file, 'w') as csvfile:
        writer = csv.writer(csvfile, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL, lineterminator = '\n')
        writer.writerow(title_row)
        for row in records:
            writer.writerow(row)

    print("Finished writing report to " + output_file)

# Following conditional statement is necessary to prevent multiprocessing from calling main multiple times on Windows
if __name__ == '__main__':
    main()
