# oyster vessel behavior classification
# Author: Avdesh Mishra

from sklearn import svm  # For SVM calculations
from sklearn import preprocessing  # For preprocessing SVM data
from sklearn.model_selection import StratifiedKFold
from sklearn.model_selection import cross_val_score
from sklearn.model_selection import GridSearchCV
from sklearn.externals import joblib  # For pickling classifiers
import os.path  # For creating filepaths
from sklearn.metrics import roc_curve, roc_auc_score, classification_report, accuracy_score, precision_score, confusion_matrix, recall_score, f1_score, auc, matthews_corrcoef
from sklearn.model_selection import train_test_split

from datetime import datetime  # For manipulating datetime objects
from time import time  # For timing of functions

def run_testing_suite(grid_searching=True, pickling=False):
    # Load the training/testing data
    curated_paths_name = os.path.join(os.path.sep, 'home', 'avdesh', 'OysterVesselBehaviorPrediction', 'OysterVesselBehaviorPred', 'code', 'all_paths_Jul13-Aug14.pkl')
    speeds, behaviors = joblib.load(curated_paths_name)
    
    # first split the dataset into validation set of 80% and independent test set of 20%
    X_val, X_test, y_val, y_test = train_test_split(speeds, behaviors, test_size=0.2, random_state=0, stratify=behaviors)
        
    
    # use validation set to perform grid search and then use the best C and gamma to obtain the 10-fold accuracy as well as build the model
    
    # Create the default scaler, then scale the speeds between -1 and 1
    scaler = preprocessing.MinMaxScaler((-1, 1))
    X_val = scaler.fit_transform(X_val, y_val)
    print("Training Set")
    print(X_val.shape[0])
    docked = 0
    fishing = 0
    travel = 0
    poling = 0
    len = y_val.shape[0]
    for i in range(len):
        if y_val[i] == 1:
            docked = docked+1
        if y_val[i] == 2:
            fishing = fishing+1
        if y_val[i] == 3:
            travel = travel+1
        if y_val[i] == 4:
            poling = poling+1
    print(docked)
    print(fishing)
    print(travel)
    print(poling)
    print("SVM")
    # Support Vector Machine, with a Radial Basis Function kernel
    if grid_searching:
        rbfsvm = svm.SVC(kernel='rbf', probability=True, random_state=144)
        param_grid = {'C': [1, 10, 100, 1000, 10000, 100000], 'gamma': [0.0001, 0.001, 0.1, 1, 10, 100, 1000]}
        optimal_params = perform_grid_search(param_grid, rbfsvm, X_val, y_val)
        rbfsvm = svm.SVC(kernel='rbf', probability=True, C=optimal_params['C'], gamma=optimal_params['gamma'], random_state=144)
    else:
        rbfsvm = svm.SVC(kernel='rbf', C=100000, gamma=1000, probability=True, random_state=144)

    # Pickle the classifier
    if pickling:
        score = str(cross_val_score(rbfsvm, X_val, y_val, cv=10, n_jobs=-1).mean())
        print("SVM 10-Fold Cross-Validation SCORE = " + score)
        rbfsvm = rbfsvm.fit(X_val, y_val)
        classifier_directory = os.path.join(os.path.sep, 'home', 'avdesh', 'OysterVesselBehaviorPrediction', 'OysterVesselBehaviorPred', 'code', 'scaled' if scaler else 'unscaled', 'rbf' + '_' + score)
        if not os.path.exists(classifier_directory):
            os.makedirs(classifier_directory)
        fp = os.path.join(classifier_directory, 'rbf_' + str(score) + '.pkl')
        joblib.dump(rbfsvm, fp)
    #    # save the scaler too, if it exists
        if scaler:
            fp = os.path.join(classifier_directory, 'scaler.pkl')
            joblib.dump(scaler, fp)
            
    # load the model file
    rbfsvm = joblib.load(os.path.join(os.path.sep, 'home', 'avdesh', 'OysterVesselBehaviorPrediction', 'OysterVesselBehaviorPred', 'code', 'scaled', 'rbf_model', 'rbf_model.pkl'))
    scaler = joblib.load(os.path.join(os.path.sep, 'home', 'avdesh', 'OysterVesselBehaviorPrediction', 'OysterVesselBehaviorPred', 'code', 'scaled', 'rbf_model', 'scaler.pkl'))
    # Now use the model created using best C and gamma value and test the model on test data
    X_test = scaler.transform(X_test)
    print("Test Set")
    print(X_test.shape[0])
    docked = 0
    fishing = 0
    travel = 0
    poling = 0
    len = y_test.shape[0]
    for i in range(len):
        if y_test[i] == 1:
            docked = docked+1
        if y_test[i] == 2:
            fishing = fishing+1
        if y_test[i] == 3:
            travel = travel+1
        if y_test[i] == 4:
            poling = poling+1
    print(docked)
    print(fishing)
    print(travel)
    print(poling)
    print("SVM")
    y_pred = rbfsvm.predict(X_test)
    class_names = ['Docked', 'Fishing', 'Travel', 'Poling']
    print(classification_report(y_test, y_pred, target_names=class_names))
    print(confusion_matrix(y_test, y_pred))
    Accuracy_cla = accuracy_score(y_test, y_pred)
    print(Accuracy_cla)
            
            
# Perform a grid search and return the best parameters
def perform_grid_search(param_grid, classifier, paths, behaviors):
    # Record the starting time
    print("Started Grid Search")
    start_time = time()
    # Use 3-fold cross validation
    skf = StratifiedKFold(n_splits=3)
    # Create the grid, and fit the data to it
    grid = GridSearchCV(classifier, param_grid=param_grid, cv=skf, n_jobs=-1)
    grid.fit(paths, behaviors)
    grid_search_time = (time() - start_time) / 60
    # Print out the results and execution time
    write_output("Classifier = %s, Time Searching = %s, Best Params = %s, Best Score = %s" % (type(classifier).__name__, grid_search_time, grid.best_params_, grid.best_score_))
    return grid.best_params_
    
    
# Method to write output of grid search to a file
def write_output(output_text):
    output_file = open('grid_search_3fold.txt', 'a')
    # Record the date and time of the output
    output_file.write('\n' + str(datetime.now()) + '\n')
    output_file.write(output_text + '\n')
    output_file.close()
    
    
def main():
    run_testing_suite(True, True)
      

    
if __name__ == '__main__':
    main()
    
    
    
    