# oyster vessel behavior classification
# Author: Avdesh Mishra

from sklearn import svm  # For SVM calculations
from sklearn import preprocessing  # For preprocessing SVM data
from sklearn.model_selection import StratifiedKFold
from sklearn.externals import joblib  # For pickling classifiers
from sklearn.metrics import roc_curve, auc  # For graphing roc curves
from sklearn.preprocessing import label_binarize  # For pre-processing ROC data
import numpy  # For ravel function
from scipy import interp
import matplotlib.pyplot as plt  # For graphing data
plt.switch_backend('agg')
import pandas as pd

import os.path  # For creating filepaths

from datetime import datetime  # For manipulating datetime objects

from time import time  # For timing of functions

#plot roc curve by performing 10-fold cross-validation
def generate_roc_curve_10fold():
    # load the original data file
    #curated_paths_name = os.path.join(os.path.sep, 'home', 'avdesh', 'OysterVesselBehaviorPrediction', 'all_paths_Jul13-Aug14.pkl')
    curated_paths_name = os.path.join(os.path.sep, 'home', 'avdesh', 'OysterVesselBehaviorPrediction', 'OysterVesselBehaviorPred', 'code', 'all_paths_Jul13-Aug14.pkl')
    speeds, behaviors = joblib.load(curated_paths_name)
    class_names = ['Docked', 'Fishing', 'Travel', 'Poling']
    #rbfsvm = svm.SVC(kernel='rbf', probability=True, C=100000, gamma=100, random_state=144)
    rbfsvm = svm.SVC(kernel='rbf', probability=True, C=10000, gamma=1000, random_state=144)
    cv = StratifiedKFold(n_splits=10)
    
    tprs_folds = []
    aucs_folds = []
    
    mean_fpr = numpy.linspace(0,1,100)
    j = 1
    for train,test in cv.split(speeds,behaviors):
        y_test = label_binarize(behaviors[test], numpy.unique(behaviors[test]))
        n_classes = y_test.shape[1]
        print(n_classes)
        print("fold_"+str(j))
        # Create the default scaler, then scale the speeds between -1 and 1
        scaler = preprocessing.MinMaxScaler((-1, 1))
        X_train = scaler.fit_transform(speeds[train])
        X_test = scaler.transform(speeds[test])
        
        prediction = rbfsvm.fit(X_train,behaviors[train]).predict_proba(X_test)
        
        tprs = []
        aucs = []
        fpr = dict()
        tpr = dict()
        roc_auc = dict()
        for i in range(n_classes):
            fpr[i], tpr[i], _ = roc_curve(y_test[:, i], prediction[:, i])
            tprs.append(interp(mean_fpr, fpr[i], tpr[i]))
            roc_auc[i] = auc(fpr[i], tpr[i])
            aucs.append(roc_auc[i])
            
        tprs_folds.append(tprs)
        aucs_folds.append(aucs)
        j = j+1
        
    df_tpr = pd.DataFrame(tprs_folds)
    df_auc = pd.DataFrame(aucs_folds)
    
    mean_tpr_docked = df_tpr[0].mean()
    mean_tpr_fishing = df_tpr[1].mean()
    mean_tpr_travel = df_tpr[2].mean()
    mean_tpr_poling = df_tpr[3].mean()
    
    mean_auc_docked = auc(mean_fpr, mean_tpr_docked)
    mean_auc_fishing = auc(mean_fpr, mean_tpr_fishing)
    mean_auc_travel = auc(mean_fpr, mean_tpr_travel)
    mean_auc_poling = auc(mean_fpr, mean_tpr_poling)
    
    fig = plt.figure(figsize=(8, 6))
    plt.plot(mean_fpr, mean_tpr_docked, color='black', lw=6, label='{0} (area = {1:0.6f})'''.format(class_names[0], mean_auc_docked))
    plt.plot(mean_fpr, mean_tpr_fishing, color='yellow', lw=3, label='{0} (area = {1:0.6f})'''.format(class_names[1], mean_auc_fishing))
    plt.plot(mean_fpr, mean_tpr_travel, color='lime', lw=2, label='{0} (area = {1:0.6f})'''.format(class_names[2], mean_auc_travel))
    plt.plot(mean_fpr, mean_tpr_poling, color='red', lw=1.5, label='{0} (area = {1:0.6f})'''.format(class_names[3], mean_auc_poling))
    #plt.plot([0, 1], [0, 1], linestyle='--')
    plt.xlim([-0.001, 0.06])
    plt.ylim([0.98, 1.01])
    #plt.xlim([-0.05, 1.05])
    #plt.ylim([-0.05, 1.05])
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.title('Mean ROC Curves for Vessel Behavior')
    plt.legend(loc="lower right")
    fig.savefig('RocPlot_oysterdata_10fold_CV_c10000_g1000_zoomed.png')
    plt.close(fig)
    
    fig1 = plt.figure()
    plt.plot(mean_fpr, mean_tpr_docked, color='black', lw=6, label='{0} (area = {1:0.6f})'''.format(class_names[0], mean_auc_docked))
    plt.plot(mean_fpr, mean_tpr_fishing, color='yellow', lw=3, label='{0} (area = {1:0.6f})'''.format(class_names[1], mean_auc_fishing))
    plt.plot(mean_fpr, mean_tpr_travel, color='lime', lw=2, label='{0} (area = {1:0.6f})'''.format(class_names[2], mean_auc_travel))
    plt.plot(mean_fpr, mean_tpr_poling, color='red', lw=1.5, label='{0} (area = {1:0.6f})'''.format(class_names[3], mean_auc_poling))
    plt.plot([0, 1], [0, 1], linestyle='--')
    #plt.xlim([-0.001, 0.06])
    #plt.ylim([0.98, 1.01])
    plt.xlim([-0.05, 1.05])
    plt.ylim([-0.05, 1.05])
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.title('Mean ROC Curves for Vessel Behavior')
    plt.legend(loc="lower right")
    fig1.savefig('RocPlot_oysterdata_10fold_CV_c10000_g1000_range_0to1.png')
    plt.close(fig1)
    
    
    
    

def main():
    generate_roc_curve_10fold()


if __name__ == '__main__':
    main()
    
